
from Canvas import Canvas
from PSMatrix import PSMatrix
import Fstr
import math	
import PSReadable
import FindResource
import re, random, os
from DeviceFont import FontTable
from Cmd import *

HexChar = ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "a", "b", "c", "d", "e", "f")
# 0 <= x <= 1
def float_to_hex(x):
	if x < 0: x = 0
	if x > 1: x = 1
	m = int(round(255*x))
	s = HexChar[m/16] + HexChar[m%16]
	return s

def pstrue(TF):
	if TF:
		return "true"
	return "false"

if True:
	savestr = ""
	restorestr = ""
else:
	savestr = "save /EPSFsave exch def "
	restorestr = "EPSFsave restore "

beginepsf = "/BeginEPSF { " + savestr + " count /OpStackSize exch def /DictStackSize countdictstack def 0 setgray 0 setlinecap 1 setlinewidth 0 setlinejoin 10 setmiterlimit [] 0 setdash newpath } bind def\n"

endepsf = "/EndEPSF { count OpStackSize sub dup 0 lt {neg {pop} repeat} {pop} ifelse countdictstack DictStackSize sub dup 0 lt {neg {end} repeat} {pop} ifelse " + restorestr + "} bind def\n"

"""
beginepsf = "/BeginEPSF { " + savestr + "(execstack) == countexecstack == count /OpStackSize exch (opstack) == dup == def /DictStackSize countdictstack (dictstack) == dup == def /showpage {} def 0 setgray 0 setlinecap 1 setlinewidth 0 setlinejoin 10 setmiterlimit [] 0 setdash newpath } bind def\n"

endepsf = "/EndEPSF { (execstack) == countexecstack == count OpStackSize (opstack) == dup == sub dup 0 lt {neg {pop} repeat} {pop} ifelse countdictstack DictStackSize (dictstack) == dup == sub dup 0 lt {neg {end} repeat} {pop} ifelse " + restorestr + "} bind def\n"
"""

"""
epsfbox = "% llx lly urx ury
/EPSFBox {
  4 dict begin
  /ury exch def
  /urx exch def
  /lly exch def
  /llx exch def
  llx lly moveto
  urx lly lineto
  urx ury lineto
  llx ury lineto
  closepath
  end
} def"
"""

class PSExec:

	""" 

	This list from Cmd.py:

	SETLINEWIDTH = 0 
	SETLINECAP = 1 
	SETLINEJOIN = 2 
	SETDASH = 3 
	SETMITERLIMIT = 4

	SETCOLOR = 5
	SCALELINEWIDTH = 6 
	NEWPATH = 7 
	MOVETO = 8 
	LINETO = 9 

	CURVETO = 10 
	FILL = 11
	STROKE = 12 
	CFILL = 13
	CSTROKE = 14

	CLIP = 15
	SETFONT = 16 
	SHOW = 17 
	EMBED = 18
	CLOSEPATH = 19 

	INSERT = 20
	IMPORTPS = 21
	IMPORTEPS = 22
	GSAVE = 23
	GRESTORE = 24

	SHFILL = 25
	COMMENT = 26
	IMAGE = 27

#	BEGINPAGE = 23
#	ENDPAGE = 24
 """

	def __init__(self):
		self.storagePath = "PiScript-tmp" + str(int(1024*random.random())) +  ".pys";
		self.storagefile = open(  self.storagePath,'w')
		self.fontTable = FontTable()
		self.pagecount = 0
		self.gsno = 0

		self.Command = [
			self.SetLineWidth,
			self.SetLineCap,
			self.SetLineJoin,
			self.SetDash,
			self.SetMiterLimit,
			#
			self.SetColor,
			self.ScaleLineWidth,
			self.Newpath,
			self.Moveto,
			self.Lineto,
			#
			self.Curveto,
			self.Fill,
			self.Stroke,
			self.CFill,
			self.CStroke,
			#
			self.Clip,
			self.SetFont,
			self.Show,
			self.Embed,
			self.Closepath,

			self.Insert,
			self.PSImport,
			self.EPSImport,
			self.GSave,
			self.GRestore,

			self.SHFill,
			self.Comment,
			self.Embed_image,
		]

	def setbbox(self, bbox):
		self.bbox = bbox

	# ----------------------------------------

	def psadd(self, ps):
		if len(self.ps) > 10000:
			self.storagefile.write( self.ps )
			self.storagefile.flush()
			self.ps = ""
		self.ps += ps 
		
	def writePage(self, insert):
		self._beginpage()
		self.fontTable.merge( insert.fontTable )
		# self.cmd = insert.cmd
		# stack = []
		self.execute(insert.cmd, [1, 0, 0, 1, 0, 0])
		self._endpage()
		self.storagefile.write( self.ps )
		self.storagefile.flush()

	def execute(self, cmd, matrix):
		n = len(cmd)
		i = 0
		while i < n:
			c = cmd[i]
			self.Command[c](cmd, i, matrix)
			i += Skip[c]

	def Insert(self, cmd, i, m):
		self.psadd("% --- raw inclusion --- \n")
		self.psadd(cmd[i+1])
		self.psadd("% --- end of inclusion --- \n")

	""" for PS only, at the moment
	img is a w x h array of triples (r,g,b) in range [0,1]
	we have to apply the current graphics state here
	because the coordinate commands translate etc. are not passed to PostScript
	"""

	def Embed_image(self, cmd, i, m):
		
		# img, n, interpolate, ctm
		img = cmd[i+1]
		n = cmd[i+2]
		interpolate = cmd[i+3]
		ctm = cmd[i+4]
		h = len(img)
		w = len(img[0])
		# print "(w,h) =", w, h
		self.psadd("% --- bit map inclusion --- \n")
		s = "/DeviceRGB setcolorspace\n"
		# apply current coordinates
		s += "[ " 
		for i in range(6):
			s += str(ctm[i]) + " "
		s += "] concat\n"
		s += "<<\n"
		s += "/ImageType 1\n"
		s += "/Width " + str(w) + "\n"
		s += "/Height " + str(h) + "\n"
		s += "/BitsPerComponent 8\n"
		s += "/Decode [ 0 1 0 1 0 1 ]\n"
		s += "/ImageMatrix [ " + str(n) + " 0 0 " + str(n) + " 0 0 ]\n"
		s += "/Interpolate " + pstrue(interpolate) + "\n"
		s += "/DataSource currentfile /ASCIIHexDecode filter\n"
		s += ">>\n"
		s += "image\n"
		# now the image
		for i in range(h):
			row = img[i]
			for j in range(w):
				(r,g,b) = row[j]
				s += float_to_hex(r)
				s += float_to_hex(g)
				s += float_to_hex(b)
			s += "\n"
		s += ">\n"
		self.psadd("% --- end of bit map inclusion --- \n")
		self.psadd(s)

	""" 
	def EPSImport(self, cmd, i, m):
		# print "*******************************************************"
		eps = cmd[i+1]
		tm = cmd[i+2]
		tm = PSMatrix.concat(tm, m)
		f = open(eps, "rb")
		F = f.read()
		f.close()
		# self.ps += self.EPSinclusion(F, tm)
		self.EPSinclude(F, tm)
	"""

	def PSImport(self, cmd, i, m):
		eps = cmd[i+1]
		tm = cmd[i+2]
		tm = PSMatrix.concat(tm, m)
		# self.psadd("gsave\n")
		self.psadd("[ ")
		for k in range(6):
			self.psadd(" " + str(tm[k]) )
		self.psadd(" ] concat\n")
		self.psadd(("%% --- inserting file " + eps + " --- \n") )
		f = open(eps, "rb")
		F = f.read()
		self.psadd(F)
		self.psadd("%% --- end of insert " + eps + " --- \n")
		# self.psadd("grestore\n")

	def epsboundingbox(filename):
		f = open(filename, "rb")
		bbox = None
		line = f.readline()
		while line: 
			m = re.search("%%BoundingBox:[ ]+([0-9]+)[ ]+([0-9]+)[ ]+([0-9]+)[ ]+([0-9]+)", line)
			if m:
				llx = int(m.group(1))
				lly = int(m.group(2))
				urx = int(m.group(3))
				ury = int(m.group(4))
				return [ llx, lly, urx, ury ]

			m = re.search("%%BoundingBox:[ ]+(\(atend\))", line)
			if m: # scan through quickly to find box at end
				depth = 0
				line = f.readline()
				m = re.search("%%Trailer", line)
				while line and (m == None or not depth == 0):
					m = re.search("%%BeginDocument", line)
					if m: 
						depth += 1
					m = re.search("%%EndDocument", line)
					if m: 
						depth -= 1
					line = f.readline()
					m = re.search("%%Trailer", line)
				# when it stops, eof [error] or a Trailer with depth = 0
			line = f.readline()
		if not line:
			print "Error: " + filename + " is not conforming EPS"
		return bbox
	epsboundingbox=staticmethod(epsboundingbox)

	"""
	def EPSinclusion(F, tm):
		ps = ("gsave\n")
		ps += ("save /SS exch def\n")
		# ps += ("1 dict begin\n") # XXX
		ps += ("/showpage{} def\n")
		ps += "gsave\n"
		ps += "[ "
		for k in range(6):
			ps += " " + str(tm[k]) 
		ps += " ] concat\n"
		ps += ("%%BeginDocument\n")
		ps += F
		ps += ("%%EndDocument\n")
		# ps += ("end\n")
		ps += ("SS restore\n")
		ps += ("grestore\n")
		return ps
	EPSinclusion=staticmethod(EPSinclusion)
	"""

	""" 
		There is a bug in `setfont' that gives grief 
		when importEPS includes a file with that command
	"""
	def EPSinclude(self, F, tm):
		
		self.psadd("gsave\n")
		self.psadd("[ ")
		for k in range(6):
			self.psadd(" " + str(tm[k]) )
		self.psadd(" ] concat\n")
		# self.psadd("save /SS exch def\n")"""
		self.psadd("1 dict begin\n")
		self.psadd(endepsf)
		self.psadd(beginepsf)
		self.psadd("/showpage{} def\n")
		self.psadd("BeginEPSF\n")
		self.psadd("%%BeginDocument:\n")
		self.psadd(F)
		self.psadd("%%EndDocument\n") 
		self.psadd("EndEPSF\n")
		self.psadd("end\n")
		# self.psadd("SS restore\n")
		self.psadd("grestore\n")

	def EPSImport(self, cmd, i, m):
		# print "*******************************************************"
		eps = cmd[i+1]
		tm = cmd[i+2]
		tm = PSMatrix.concat(tm, m)
		f = open(eps, "rb")
		F = f.read()
		f.close()
		# self.ps += self.EPSinclusion(F, tm)
		self.EPSinclude(F, tm)

	def _beginpage(self):
		self.gsno = 0
		self.pagecount += 1
		pageno = self.pagecount
		self.ps = ("%%Page: " + str(pageno) + " " + str(pageno) + "\n")	

	def _endpage(self):
		for i in range(self.gsno):
			self.psadd("grestore ")
		self.psadd("showpage\n")
		# self.psadd("%%EndPage\n")
		self.psadd("\n% ---------------------------------------------------------\n\n")

	def GSave(self, cmd, i, m):
		self.gsno += 1
		self.psadd("gsave\n")

	def GRestore(self, cmd, i, m):
		self.gsno -= 1
		self.psadd("grestore\n")
		
	def Embed(self, cmd, i, m): 
		c = cmd[i+1] # a Canvas
		m = PSMatrix.concat(m, c.m)
		# print "PSExec Embed: matrix = ", m
		self.execute(c.cmd, m)

	def SetLineWidth(self, cmd, i, m):
		ps = Fstr.fstr(cmd[i+1])
		ps += " setlinewidth\n"
		self.psadd(ps)

	def SetLineCap(self, cmd, i, m):
		ps = str(cmd[i+1])
		ps += " setlinecap\n"
		self.psadd(ps)

	def SetLineJoin(self, cmd, i, m):
		ps = str(cmd[i+1])
		ps += " setlinejoin\n"
		self.psadd(ps)

	def topsarray(a):
		s = "["
		for x in a:
			s += " " + str(x)
		s += " ]"
		return s

	topsarray = staticmethod(topsarray)

	def SetDash(self, cmd, i, m):
		a = cmd[i+1]
		o = cmd[i+2]
		ps = self.topsarray(a) + " " + str(o)
		ps += " setdash\n"
		self.psadd(ps)

	def SetColor(self, cmd, i, m):
		ps = Fstr.fstr(cmd[i+1])
		ps += " " + Fstr.fstr(cmd[i+2])
		ps += " " + Fstr.fstr(cmd[i+3])
		ps += " setrgbcolor\n"
		self.psadd(ps)

	def ScaleLineWidth(self, cmd, i, m):
		ps = "currentlinewidth "
		ps += Fstr.fstr(cmd[i+1])
		ps += " mul setlinewidth\n"
		self.psadd(ps)

	def Newpath(self, cmd, i, m):
		self.psadd("newpath\n")

	def Moveto(self, cmd, i, m): # gfx
		( x, y ) = ( cmd[i+1], cmd[i+2] )
		( x, y ) = PSMatrix.transform(m, x, y)
		# print "PSExec Moveto: ", (x, y)
		ps = Fstr.fstr(x)
		ps += " " + Fstr.fstr(y)
		ps += " moveto\n"
		self.psadd(ps)

	def Lineto(self, cmd, i, m): # gfx
		( x, y ) = ( cmd[i+1], cmd[i+2] )
		( x, y ) = PSMatrix.transform(m, x, y)
		ps = Fstr.fstr(x)
		ps += " " + Fstr.fstr(y)
		ps += " lineto\n"
		self.psadd(ps)

	def Curveto(self, cmd, i, m): # gfx
		( x, y ) = ( cmd[i+1], cmd[i+2] )
		( x, y ) = PSMatrix.transform(m, x, y)
		ps = Fstr.fstr(x)
		ps += " " + Fstr.fstr(y)

		( x, y ) = ( cmd[i+3], cmd[i+4] )
		( x, y ) = PSMatrix.transform(m, x, y)
		ps += " " + Fstr.fstr(x)
		ps += " " + Fstr.fstr(y)

		( x, y ) = ( cmd[i+5], cmd[i+6] )
		( x, y ) = PSMatrix.transform(m, x, y)
		ps += " " + Fstr.fstr(x)
		ps += " " + Fstr.fstr(y)

		ps += " curveto\n"
		self.psadd(ps)

	def Fill(self, cmd, i, m):
		self.psadd("gsave fill grestore\n")

	def Stroke(self, cmd, i, m):
		self.psadd("gsave stroke grestore\n")

	def CFill(self, cmd, i, m):
		ps = "gsave\n"
		ps += Fstr.fstr(cmd[i+1])
		ps += " " + Fstr.fstr(cmd[i+2])
		ps += " " + Fstr.fstr(cmd[i+3])
		ps += " setrgbcolor\n"
		ps += "fill\n"
		ps += "grestore\n"
		self.psadd(ps)

	def CStroke(self, cmd, i, m):
		ps = "gsave\n"
		ps += Fstr.fstr(cmd[i+1])
		ps += " " + Fstr.fstr(cmd[i+2])
		ps += " " + Fstr.fstr(cmd[i+3])
		ps += " setrgbcolor\n"
		ps += "stroke\n"
		ps += "grestore\n"
		self.psadd(ps)

	def Clip(self, cmd, i, m):
		self.psadd("clip\n")

	def SetFont(self, cmd, i, m):
		# The font name is for internal PS use
		# the font has alraedy been loaded into the PS file
		ps = "/" + cmd[i+1]
		ps += " findfont\n"
		ps += str(cmd[i+2])
		ps += " scalefont\n"
		ps += "setfont\n"
		self.psadd(ps)

	def SetMiterLimit(self, cmd, i, m):
		ps = Fstr.fstr(cmd[i+1])
		ps += " setmiterlimit\n"
		self.psadd(ps)

	# WAC, FIXME: rewrite the string to conform to ascii
	# this is the only place a coordinate change is made in PostScript
	# show string matrix OR (new) int matrix
	# NEW: cmd[i+1] = array of integers (so we don'thave to convert to ord(c))
	def Show(self, cmd, i, m):
		ps = ""
		tm = cmd[i+2]
		tm = PSMatrix.concat(tm, m)
		# ps += "[ currentpoint ] == \n"
		ps += "gsave\n"
		ps += "[ "
		for k in range(6):
			ps += " " + str(tm[k]) 
		ps += " ] concat\n"
		outString = ""
		for c in cmd[i+1]:
			outString += PSReadable.toReadable[c]
		ps += "(" + outString + ")"
		ps += " show\n"
		ps += "grestore\n"
		self.psadd(ps)

	def Closepath(self, cmd, i, m):
		self.psadd("closepath\n")

	def SHFill(self, cmd, i, m):
		t = cmd[i+1]
		ds = cmd[i+2]
		if t == 6:
			self.psadd("\n")
			self.psadd("newpath\n")
			self.psadd("<<\n")
			self.psadd("/ShadingType 6\n")
  			self.psadd("/ColorSpace [ /DeviceRGB ]\n")
  			self.psadd("/DataSource [\n")
			# ds = 12 pairs x y + 4 colours r g b
			s = "0"
			for x in ds:
				s += " "
				s += Fstr.fstr(x)
			s += " \n"
			self.psadd(s)
			self.psadd("]\n")
			self.psadd(">>\n")
			self.psadd("shfill\n\n")
		elif t == 4:
			self.psadd("\n")
			self.psadd("newpath\n")
			self.psadd("<<\n")
			self.psadd("/ShadingType 4\n")
  			self.psadd("/ColorSpace [ /DeviceRGB ]\n")
  			self.psadd("/DataSource [\n")
			for t in ds:
				for D in t:
					P = D[0]
					C = D[1]
					self.psadd("0 " + Fstr.fstr(P[0]) + " " + Fstr.fstr(P[1]) + " " + Fstr.fstr(C[0]) + " " + Fstr.fstr(C[1]) + " " + Fstr.fstr(C[2]) + "\n")
			self.psadd("]\n")
			self.psadd(">>\n")
			self.psadd("shfill\n\n")
		else: # t == 2
			x0 = ds[0]
			y0 = ds[1]
			x1 = ds[2]
			y1 = ds[3]
			self.psadd("\n")
			self.psadd("newpath\n")
			self.psadd("<<\n")
			self.psadd("/ShadingType 2\n")
  			self.psadd("/ColorSpace /DeviceRGB\n")
  			self.psadd("/Coords [ " + Fstr.fstr(x1) + " " + Fstr.fstr(y1) + " " + Fstr.fstr(x0) + " " + Fstr.fstr(y0) + " ]\n")
  			self.psadd("/Function <<\n")
			self.psadd("\t/FunctionType 2\n")
			self.psadd("\t/Domain [ 0 1 ]\n")
			C0 = [ ds[7], ds[8], ds[9] ]
			self.psadd("\t/C0 [ " + Fstr.fstr(C0[0]) + " " + Fstr.fstr(C0[1]) + " "  + Fstr.fstr(C0[2]) + " ]\n") 
			C1 = [ ds[4], ds[5], ds[6] ]
			self.psadd("\t/C1 [ " + Fstr.fstr(C1[0]) + " " + Fstr.fstr(C1[1]) + " "  + Fstr.fstr(C1[2]) + " ]\n") 
			self.psadd("/N 1\n")
			self.psadd(">>\n")
			self.psadd(">>\n")
			self.psadd("shfill\n\n")
			

	def Comment(self, cmd, i, m):
		comment = cmd[i+1]
		self.psadd("\n% Comment: " + comment + "\n\n")

	def roundBBox( a, b ):
		import math
		if a < b:
			a = math.floor(a); b = math.ceil(b)
		else:
			a = math.ceil(a); b = math.floor(b)
		return (a,b)
	roundBBox = staticmethod(roundBBox)

	def finish(self, outFile, toEPS=True ):

		self.storagefile.close()
		self.storagefile = open(self.storagePath, "r")

		if toEPS:
			self.header = "%!PS-Adobe-2.0 EPSF-3.0\n"
		else:
			self.header = "%!PS-Adobe-2.0\n"
		[ llx, lly, urx, ury ] = self.bbox
		(llx, urx) = self.roundBBox( llx, urx )
		(lly, ury) = self.roundBBox( lly, ury )
		self.header += ("%%%%BoundingBox: %d %d %d %d\n" % ( llx, lly, urx, ury) )

		self.header += "%%Pages: " + str(self.pagecount) + "\n"
		self.header += "%%PageOrder: Ascend\n\n"
		self.header += "%%BeginProlog\n"

		self.header += "%%BeginProcSet: \n"
		self.header += "/ReEncodeFont { exch findfont << >> copy dup 3 2 roll /Encoding exch put definefont } def\n"
		self.header += "%%EndProcSet\n"
		outFile.write( self.header )

		# For each raw device font used, track the set of glyphs used for that font.
		for deviceFont in self.fontTable.deviceFonts.values():
			deviceFont.glyphsUsed = deviceFont.encodingVector().glyphSetForChars( deviceFont.charsUsed() )

		# For each encoded font, add the set of glyphs used to its deviceFont
		for encodedFont in self.fontTable.encodedFonts.values():
			encodedGlyphsUsed = encodedFont.encodingVector().glyphSetForChars( encodedFont.charsUsed() )
			encodedFont.deviceFont.glyphsUsed.update( encodedGlyphsUsed )

		# Determine a list of unique encoding vectors
		encodings = set()
		for encFont in self.fontTable.encodedFonts.values():
			if encFont.encodingFile:
				encodings.add( encFont.encodingFile )

		encodingNames = {}
		# For each unique encoding file, put a copy in the output file
		for e in encodings:
			encodingNames[e] = self.emitEncoding( e, outFile )

		for f in self.fontTable.deviceFonts.values():
			if (not f.shouldEmbed) or (not f.fontPath):
				continue
			# logging.debug("processing font %s with %d characters" % (str(f), len(f.glyphsUsed)) )
			# If we have a physical path, we assume it's a Type1Font.
			# FIXME (DM 07/09/09): What if the constructor of Type1Font fails?  I think we'll crash.
			t = FindResource.type1FontForPath( f.fontPath )
			if f.shouldExtract:
				fout = t.extractPFA( f.glyphsUsed )
			else:
				fout = t.toPFA()
			outFile.write("%%BeginFont: " + f.PSName + "\n")
			outFile.write( fout )
			outFile.write( "%%EndFont " + f.PSName + "\n" )

		outFile.write("%%EndProlog\n")

		for f in self.fontTable.encodedFonts.values():
			if( f.encodingFile == None ):
				continue
			fullEncodingName = encodingNames[f.encodingFile]				
			if( fullEncodingName == None ):
				logging.warning("Unable to find the encoding map name for encoding  %s", f.encoding)
				continue
			outFile.write( "/%s /%s %s ReEncodeFont\n" % (f.uniqueName(), f.deviceFont.PSName, fullEncodingName) )
	
		outFile.flush()

		# read tmp output, write it
		f = self.storagefile.read()
		outFile.write( f )
		self.storagefile.close()
		os.remove( self.storagePath )

		outFile.write("%%Trailer\n")
		outFile.write("%%EOF\n")

	def emitEncoding( self, encodingFile, out ):
		import re
		
		encPath = FindResource.getEncoding(encodingFile )
		if( encPath == None ):
			logging.warning("Unable to find encoding file %s", encodingFile )
			return None
			# The postscript is going to be mangled now. We've been using a font name
			# for which there is now no definition.  Should FIXME by adding code in
			# the PS to remap the encoded font name back to the original font name
			# when we defined the font encodings. Or maybe just abort?

		encFile = open(encPath, "rb")

		header = "%%%%BeginProcSet: %s 0 0\n" % encodingFile
		out.write(header)

		encName = None
		name_re=re.compile("^\s*/(\w+)\s.*")
		for l in encFile:
			m = name_re.match(l)
			if( m ):
				encName = m.group(1)
				break
		encFile.seek(0)
		for l in encFile:
			out.write(l)
		out.write("%%EndProcSet\n")
		return encName



# =================================================================

