"""
	Each Canvas has its own coordinate system
		+ boundary and origin 

	TexInserts have special points called marks.
"""


from DviToDevice import DviToDevice
"""
WAC: none of these used here:
from Graphics import Graphics
from PSMatrix import PSMatrix
from Canvas import Canvas 
from CoordinateSystem import CoordinateSystem """

import os
import random 

# ---------------------------------------------------------

class TexRunner:
	def __init__(self, texenv, texString, device, save=None, pin=False):
		self.texenv = texenv
		self.device = device
		self.string = texString
		self.label = save
		self.pinned = pin 
		self.run()

	def run(self):
		# print "running ... "
		# add prefix, postfix to texstring, write to a tex file
		# tex that file, apply DviEPS to that .dvi file
		if self.label == None:
			n = int(1024*random.random())
			filename = "tmp" + str(n)
		else:
			filename = self.label
		self.filename = filename
		contents = self.texenv.prefix
		contents += "\n" + self.texenv.macros
		contents += "\n" + self.string
		contents += "\n" + self.texenv.postfix
		# print contents
		# --- run tex to produce a .dvi file -------------------
		if self.label == None:
			# print "TEXING"
			r = self.tex(filename, contents)
		else:
			# check to see if the tex file exists and has changed
			if os.path.exists(filename + ".tex") and os.path.exists(filename + ".dvi"):
				texin = open(filename + ".tex", "r")
				oldcontents = texin.read()
				texin.close()
				if not contents == oldcontents:
					# run tex on new tex file: filename, contents = arguments
					r = self.tex(filename, contents)
				else:
					pass
					# print "Reusing old dvi file!"
			else:
				# print "texing "
				self.tex(filename, contents)
		# --- read the .dvi file into the device
		dvitodevice = DviToDevice( self.filename, self.device )
		dvitodevice.render()	

		if self.label == None: # cleanup
			self.cleanup()
			# pass
		
	def cleanup(self):
		filename = self.filename
		if self.label == None: # cleanup
			if (os.path.exists(filename + ".tex")):
				os.remove(filename + ".tex") 
			if (os.path.exists(filename + ".dvi")):
				os.remove(filename + ".dvi")
				
		if (os.path.exists(filename + ".log")):
			os.remove(filename + ".log")
		if (os.path.exists(filename + ".aux")):
			os.remove(filename + ".aux")

	def tex(self, filename, contents):
		# print "TEX command: " + "\"" + self.texenv.command + "\""
		texout = open(filename + ".tex", "w")
		texout.write(contents)
		texout.close() 
		cmd = self.texenv.command + " " + filename
		print "TEX cmd =", cmd
		R = os.system(cmd)
		# O = os.system("ls")
		# print "ls:", O
		return R


	# ------------------------------------------------------------------------

class TexEnv:
	def __init__(self, p, q, c):
		self.prefix = p
		self.macros = ""
		self.postfix = q
		self.command = c
		self.save = False

	def setprefix(self, s):
		self.prefix = s

	def setmacros(self, s):
		self.macros = s

	def setpostfix(self, s):
		self.postfix = s

	def setcommand(self, s):
		self.command = s

	# T/F only
	def setsave(self, s):
		self.save = s



