
from Fstr import *
import math

class Vector:
	# v = array of numbers
	def __init__(self, *args):
		if self.isarray(args[0]):
			v = args[0]
			u = []
			for i in range(len(v)):
				u.append(v[i])
			self.v = u
		else:
			n = len(args)
			v = []
			for i in range(n):
				v.append(args[i])
			self.v = v
			
	def clone(self):
		v = self.v
		u = []
		for i in range(len(v)):
			u.append(v[i])
		return Vector(u)

	def __len__(self):
		return len(self.v)

	def __neg__(self):
		u = self.v
		v = []
		n = len(u)
		for i in range(n):
			v.append(-u[i])
		return Vector(v)

	# dot product
	def __mul__(self, v):
		if isinstance(v, Vector) or type(v) == list or type(v) == tuple:
			u = self.v
			n = len(u)
			p = u[0]*v[0]
			for i in range(1, n):
				p += u[i]*v[i]
			return p
		else: # V = scale
			u = self.v
			n = len(u)
			V = []
			for i in range(n):
				V.append (v*u[i])
			return Vector(V)

	# scalar division by c
	def __div__(self, c):
		u = self.v
		n = len(u)
		v = []
		for i in range(n):
			v.append (u[i]/c)
		return Vector(v)

	# sum called by +
	def __add__(self, v):
		u = self.v
		n = len(u)
		p = []
		for i in range(0, n):
			p.append(u[i] + v[i])
		return Vector(p)

	def __radd__(self, v):
		u = self.v
		n = len(u)
		p = []
		for i in range(0, n):
			p.append(u[i] + v[i])
		return Vector(p)

	# sum called by +
	def __sub__(self, v):
		u = self.v
		n = len(u)
		p = []
		for i in range(0, n):
			p.append(u[i] - v[i])
		return Vector(p)

	# scalar product
	def __imul__(self, c):
		# c = scale
		u = self.v
		n = len(u)
		for i in range(n):
			u[i] *= c
		return self

	# scalar product
	def __rmul__(self, c):
		# c = scale
		v = self.v
		n = len(v)
		u = [None]*n
		for i in range(n):
			u[i] = c*v[i]
		return Vector(u)

	# scalar division by c
	def __idiv__(self, c):
		# c = scale
		u = self.v
		n = len(u)
		V = []
		for i in range(n):
			u[i] /= c
		return self

	# sum called by +
	def __iadd__(self, v):
		u = self.v
		n = len(u)
		p = []
		for i in range(0, n):
			u[i] += v[i]
		return self

	# sum called by +
	def __isub__(self, v):
		u = self.v
		n = len(u)
		p = []
		for i in range(0, n):
			u[i] -= v[i]
		return self

	def __abs__(self):
		r = self*self
		return math.sqrt(r)

	def __getitem__(self, i):
		return self.v[i]

	def __setitem__(self, i, x):
		self.v[i] = x

	def __cmp__(self, v):
		u = self.v
		n = len(u)
		for i in range(n):
			if u[i] < v[i]:
				return -1
			if u[i] > v[i]:
				return 1
		return 0

	def dim(self):
		return len(self.v)

	"""	[ u0 u1 u2 ]
		[ v0 v1 v2 ] """
	def x(self, v):
		u = self.v
		w = []
		w.append(u[1]*v[2]-u[2]*v[1])
		w.append(u[2]*v[0]-u[0]*v[2])
		w.append(u[0]*v[1]-u[1]*v[0])
		return Vector(w)

	def cross(self, v):
		u = self.v
		w = []
		w.append(u[1]*v[2]-u[2]*v[1])
		w.append(u[2]*v[0]-u[0]*v[2])
		w.append(u[0]*v[1]-u[1]*v[0])
		return Vector(w)

	def normalize(self):
		r = abs(self)
		for i in range(len(self.v)):
			self.v[i] /= r
		return self

	def normalized(self):
		r = abs(self)
		return self/r

	# only for 2D vectors
	def arg(self):
		return math.atan2(self.v[1], self.v[0])
	
	def perp(self):
		v = self.v
		return Vector([-v[1], v[0]])

	def rotated(self, a):
		c = math.cos(a); s = math.sin(a)
		v = self.v
		return Vector([ c*v[0] - s*v[1], s*v[0] + c*v[1] ])

	def length(self):
		r = self*self
		return math.sqrt(r)

	def linethrough(self, Q):
		P = self.v
		v = [ Q[0]-P[0], Q[1]-P[1] ]
		r = length(v)
		A = -v[1]/r
		B = v[0]/r
		C = A*P[0]+B*P[1]
		return [ A, B, -C ]

	# called by print
	def __str__(self):
		s = "["
		u = self.v
		n = len(u)
		for i in range(n):
			s += " " + fstr(u[i])
			if i < n-1:
				s += ","
		s += " ]"
		return(s)

	def rotate(self, A):
		c = math.cos(A); s = math.sin(A)
		u = self.v
		x = c*u[0]-s*u[1]; y = s*u[0]+c*u[1] 
		u[0] = x; u[1] = y

	def interpolated(self, q, t):
		p = self.v
		s = 1-t
		r = [ ]
		for i in range(len(p)):
			r.append(s*p[i]+t*q[i])
		return Vector(r)

	def angle_to(self, v):
		u = self
		a = abs(u)
		b = abs(v)
		c = u*v
		# not a great formula, does not deal with rounding errors well
		theta = math.acos(c/(a*b))
		return theta
	
	# --- static utilities -----------------------------------------

	def isarray(a):
		if "__getitem__" in dir(a):
			return True
		return False
	isarray=staticmethod(isarray)

# ---------------------------------------------

"""
u = Vector([0,0])
v = Vector([1,0])
w = Vector([0,1])

print u < w

"""

